@@ -1,45 +1,44 @@
 ## Author: Patrick Miller
 ## Purpose: Multiple response (or multivariate) boosting with decision trees (stumps). Also multi-task, or multi-objective. Only for continuous Y.
-## RRV: 5/21/2015
+## RRV: 9/3/2015
   
 
 #' Fitting a Multivariate Tree Boosting Model
 #'
 #' Builds on \code{gbm} (Ridgeway 2013; Friedman, 2001) to fit a univariate tree model for each outcome, selecting predictors at each iteration that explain covariance between the outcomes. The number of trees included in the model can be chosen by minimizing the multivariate mean squared error using cross validation or a test set.
 #'
-#' @param X matrix or data.frame of predictors. For best performance, continuous predictors should be scaled. Categorical variables should be factors.
-#' @param Y vector, matrix, or data.frame for outcome variables. For best performance, outcome variables should be scaled.
-#' @param n.trees maximum number of iterations (trees) to be included in the model
-#' @param shrinkage amount of shrinkage to be applied in each univariate model. This is the gbm shrinkage. Default is .01. Smaller shrinkage values require more iterations to ensure good fit.
-#' @param interaction.depth fixed depth of trees to be included in the model. A tree depth of 1 are stumps (main effects only), higher tree depths capture higher order interactions.
-#' @param bag.frac   proportion of the training sample used to fit univariate trees for each response at each iteration. Default: .5
-#' @param cv.folds   number of cross validation folds. Default: 1. For larger values, runs k + 1 models, where the k models are run in parallel.
-#' @param trainfrac  proportion of the sample used for training the multivariate additive model. If both cv.folds and trainfrac are specified, the CV is done within the training set only.
-#' @param samp.iter  Experimental, does nothing for now.
+#' @param X vector, matrix, or data.frame of predictors. For best performance, continuous predictors should be scaled to have unit variance. Categorical variables should converted to factors.
+#' @param Y vector, matrix, or data.frame for outcome variables. Missing values must be imputed. To easily compare influences across outcomes and for numerical stability, outcome variables should be scaled to have unit variance.
+#' @param n.trees maximum number of trees to be included in the model. Trees are grown until a minimum number observations in each node is reached. This minimum can be modified using additional arguments (below).
+#' @param shrinkage a constant multiplier for the predictions from each tree to ensure a slow learning rate. Default is .01. Small shrinkage values may require a large number of trees to provide adequate fit.
+#' @param interaction.depth fixed depth of trees to be included in the model. A tree depth of 1 corresponds to fitting stumps (main effects only), higher tree depths capture higher order interactions.
+#' @param bag.frac   proportion of the training sample used to fit univariate trees for each response at each iteration. Default: 1
+#' @param cv.folds   number of cross validation folds. Default: 1. Runs k + 1 models, where the k models are run in parallel and the final model is run on the entire sample. If larger than 1, the number of trees that minimize the multivariate MSE averaged over k-folds is reported in object$best.trees.
+#' @param trainfrac  proportion of the sample used for training the multivariate additive model. The number of trees that minimize the multivariate MSE in the test set are given in out$best.trees. If both cv.folds and trainfrac are specified, the CV is done within the training set only.
 #' @param s vector of indices denoting observations to be used for the training sample. If s is specified, trainfrac is ignored.
-#' @param seednum integer < 1000 to ensure that results are reproducible
+#' @param seednum integer to ensure that results are reproducible
 #' @param compress T/F. Compress output results list using bzip2 (approx 10\% of original size). Default FALSE.
-#' @param save.cv  T/F. Save all k cv models.
+#' @param save.cv  T/F. Save all k-fold cross-validation models.
 #' @param mc.cores Number of cores for cross validation.
 #' @param alpha optional argument to select predictors that explain more variance or covariance in outcomes. Variance reductions are weighted by alpha, and covariance reductions are weighted by 1-alpha.
-#' @param weight.type see Details.
-#' @param cov.discrep see Details.
-#' @param ... additional arguments passed to \code{gbm}.
+#' @param weight.type Experimental.
+#' @param cov.discrep Experimental. Choose the type of covariance discrepancy.
+#' @param samp.iter   Experimental.
+#' @param ... additional arguments passed to \code{gbm}. These include \code{distribution}, \code{weights}, \code{var.monotone}, \code{n.minobsinnode}, \code{keep.data}, \code{verbose}, \code{class.stratify.cv}.  Note that other \code{distribution} arguments have not been tested.
 #' @return Fitted model. This is a list containing the following elements:
 #' 
 #' \itemize{
-#'   \item models - list of gbm models for each outcome.
-#'   \item covex - covariance explained in each pair of outcomes by each predictor, in original order.
-#'   \item maxiter - maximum number of iterations run
-#'   \item best.trees - list of the best best iteratino (min MSE test, min MSE cv, last)
+#'   \item models - list of gbm models for each outcome. Functions from the gbm package (e.g. to compute relative influence, print trees, obtain predictions, etc) can be directly applied to each of these models 
+#'   \item covex - covariance explained in each pair of outcomes by each predictor. The covariance explained is only unambiguous if predictors are independent, otherwise it is an approximation. If the interaction.depth is larger than 1, the covariance explained is attributed to the predictor with the largest effect.
+#'   \item maxiter - maximum number of iterations run (the number of trees fit)
+#'   \item best.trees - list of the best number of trees given by minimizing the multivariate MSE error in the test set, by CV, or just the last tree fit. Many of the functions in the package default to using the minimum value of the three. 
 #'   \item rel.infl - n x q x n.trees matrix of relative influences
-#'   \item w.rel.infl - n x q x n.trees matrix of weighted relative influences
+#'   \item w.rel.infl - n x q x n.trees matrix of weighted relative influences (see details)
 #'   \item params - arguments to mvtb
-#'   \item trainerr - multivariate training error at each tree
+#'   \item trainerr - multivariate training error at each tree.
 #'   \item testerr  - multivariate test error at each tree (if trainfrac < 1)
 #'   \item cverr    - multivariate cv error at each tree (if cv.folds > 1)
-#'   \item bestxs - vector of predictors selected at each tree
-#'   \item bestys - vector of dependent variables selected at each tree
+#'   \item bestxs - matrix of predictors selected at each tree
 #'   \item resid - n x q matrix of residuals after fitting all trees
 #'   \item ocv - if save.cv=TRUE, returns the CV models.
 #'   \item wm.raw - raw decreases in covariance attributable to a given tree
@@ -50,10 +49,44 @@
 #'   \item ynames
 #' }
 #' 
-#' @details Note that this is a beta version with details subject to change. Any contributions are welcome.
-#' @seealso \code{summary.mvtb}, \code{predict.mvtb}, \code{mvtb.nonlin} to help detect nonlinear effects, \code{plot.mvtb}, \code{mvtb.perspec} for plots, \code{cluster.covex}, \code{heat.covex} to compute the covariance in the outcomes explained by predictors and plot in a heatmap
-#' \code{uncomp.mvtb} to uncompress a compressed output object
-#' @references Miller P.J., Lubke G.H, McArtor D.B., Bergeman C.S. (Submitted) Finding structure in data: A data mining alternative to multivariate multiple regression. Psychological Methods.
+#' @details 
+#' 
+#' This function computes separate gbm models for each outcome (contained in \code{$models}). From these models, the covariance explained by 
+#' each predictor is then computed based on the reduction in covariance between outcomes that results from fitting a single tree to each outcome, one outcome at a time.
+#' The reduction in covariance between each pair of outcomes due to splitting on each predictor over all trees is the 'covariance explained' by each predictor, and is recorded in \code{$covex}.
+#' 
+#' The rows (pairs of outcomes) and the columns (predictors) of \code{$covex} can be clustered so that groups of predictors that explain similar pairs of covariances are closer together (see  \code{mvtb.cluster}). 
+#' A simple heatmap of this matrix can be obtained by the function \code{mvtb.heat}. The \code{covex} by each predictor is only unambiguous if the predictors are uncorrelated and interaction.depth = 1. 
+#' If predictors are not independent, the decomposition of covariance explained is only approximate (like the decomposition of R^2 by each predictor in a linear model). 
+#' If interaction.depth > 1, the following heuristic is used: the covariance explained by the tree is assigned to the predictor with the largest influence in each tree.
+#' 
+#' (Relative) influences can be retrieved using \code{mvtb.ri}, which are the usual reductions in SSE due to splitting on each predictor. 'Weighted' influences for each predictor are the usual reductions in SSE weighted by
+#' the covariance explained in all pairs of outcomes by that predictor. This allows predictor selection to be informed by the covariance explained. Higher weight can be given to explaining variances or covariances by controlling the parameter
+#' \code{alpha}:  weight = \code{alpha}(varex) + (1-\code{alpha})(covex). By default, \code{alpha} = .5, equally weighting variance and covariance explained. 
+#' Setting \code{alpha} = 0 corresponds to weighting covariance explained only,
+#' letting \code{alpha} = 1 corresponds to weighting variance explained only.
+#' 
+#' The model is tuned jointly by selecting the number of trees that minimize multivariate mean squared error in a test set (by setting \code{trainfrac}) or averaged over k folds in k-fold cross-validation (by setting \code{cv.folds > 1}).
+#' The best number of trees is available via \code{$best.trees}. Cross-validation can be parallelized by setting mc.cores > 1.  
+#' If both \code{cv.folds} and \code{trainfrac} is specified, cross-validation is carried out within the training set.
+#' Cross-validation models are usually discarded but can be saved by setting \code{save.cv = TRUE}. CV models can be accessed from \code{$ocv} of the 
+#' output object. 
+#' Multivariate mean squared training, test, and cv error are available from \code{$trainerr, $testerr, $cverr} from the output object.
+#' Observations can be specifically set for inclusion in the training set by passing a vector of integers indexing the rows to include to \code{s}.
+#' If \code{s} is specified, \code{trainfrac} is ignored but cross-validation will be carried out for observations in \code{s}.
+#' 
+#' Since the output objects can be large, automatic compression is available by setting \code{compress=TRUE}. 
+#' All methods that use the \code{mvtb} object automatically uncompress this object if necessary. 
+#' The function \code{mvtb.uncomp} is available to manually decompress the object.
+#' 
+#' Note that trees are grown until a minimum number of observations in each node is reached. 
+#' If the number of training samples*bag.fraction is less the minimum number of observations, (which can occur with small data sets), this will throw an error. 
+#' Adjust the \code{n.minobsinnode}, \code{trainfrac}, or \code{bag.fraction}.
+#' 
+#' This is a beta version with details subject to change. Any contributions are welcome.
+#' @seealso \code{summary.mvtb}, \code{predict.mvtb}, \code{mvtb.nonlin} to help detect nonlinear effects, \code{plot.mvtb}, \code{mvtb.perspec} for plots, \code{mvtb.cluster}, \code{mvtb.heat} 
+#' \code{mvtb.uncomp} to uncompress a compressed output object
+#' @references Miller P.J., Lubke G.H, McArtor D.B., Bergeman C.S. (Submitted) Finding structure in data with multivariate tree boosting.
 #' 
 #' Ridgeway, G., Southworth, M. H., & RUnit, S. (2013). Package 'gbm'. Viitattu, 10, 2013.
 #'
@@ -63,15 +96,13 @@
 #' @examples
 #' set.seed(123)
 #' n <- 1000
+#' X <- matrix(rbinom(n*3,size=1,prob=.5),n,3)    # create 3 dichotomous predictors
+#' X2 <- cbind(x1x2=X[,1]*X[,2],x2x3=X[,2]*X[,3]) # create 2 interaction terms
+#' Xf <- cbind(X,X2)                              # full design matrix, used for data generation
+#' E <- matrix(rnorm(n*4),nrow=n,ncol=4)          # independent errors
 #' B <- matrix(0,nrow=5,ncol=4)
-#' B[3,1:2] <- 0
-#' B[2,2:3] <- 0
-#' B[4,1] <- 1
-#' B[5,3:4] <- 1
-#' X <- matrix(rbinom(n*(nrow(B)-2),size=1,prob=.5),n,nrow(B)-2)
-#' X2 <- cbind(x1x2=X[,1]*X[,2],x2x3=X[,2]*X[,3])
-#' Xf <- cbind(X,X2)
-#' E <- matrix(rnorm(n*4),nrow=n,ncol=4)
+#' B[4,1] <- 1      # x1x2 interaction has a true effect on outcome 1
+#' B[5,3:4] <- 1    # x2x3 interaction has a true effect on outcomes 3 and 4
 #' Y <- Xf %*% B + E
 #' 
 #' 
@@ -88,23 +119,25 @@
 #' plot(out)
 #' mvtb.nonlin(out,X=X,Y=Y)
 #' mvtb.perspec(out)
-#' cluster.covex(out)
-#' heat.covex(out)
+#' mvtb.cluster(out)
+#' mvtb.heat(out)
 
 #' @export
 #' @importFrom stats cov
-mvtb <- function(X=X,Y=Y,n.trees=100,shrinkage=.01,interaction.depth=1,
-                 trainfrac=1,samp.iter=FALSE,bag.frac=1,cv.folds=1,
-                 s=NULL,seednum=NULL,compress=FALSE,save.cv=FALSE,mc.cores=1,alpha=.5,cov.discrep=1,weight.type=1,...) {
+mvtb <- function(X,Y,n.trees=100,shrinkage=.01,interaction.depth=1,
+                 trainfrac=1,bag.frac=1,cv.folds=1,
+                 s=NULL,seednum=NULL,compress=FALSE,save.cv=FALSE,mc.cores=1,samp.iter=FALSE,alpha=.5,cov.discrep=1,weight.type=1,...) {
 
-  Y <- as.matrix(Y)
+  if(class(Y) != "matrix") { Y <- as.matrix(Y) }
+  if(is.null(ncol(X))){ X <- as.matrix(X)}
+  #if(class(X) != "matrix") { X <- data.matrix(X) }
   params <- c(as.list(environment()))
   ## seeds
   if(!is.null(seednum)){
     #print(c("mvtboost: seednum ",seednum));
     set.seed(seednum)
   }
- 
+  stopifnot(nrow(X) == nrow(Y))
   
   ## Data
   n <- nrow(X); k <- ncol(Y); p <- ncol(X);
@@ -143,7 +176,8 @@ mvtb <- function(X=X,Y=Y,n.trees=100,shrinkage=.01,interaction.depth=1,
   rel.infl <- w.rel.infl <- array(0,dim=c(p,k,n.trees)) # influences at every iteration
   
   ## Iterations
-  trainerr <- testerr <- bestys <- bestxs <- vector(length=n.trees)
+  trainerr <- testerr <- vector(length=n.trees)
+  bestxs   <- matrix(0,nrow=n.trees,ncol=k) 
   final.iter <- FALSE
   
   ## Covex
@@ -190,9 +224,10 @@ mvtb <- function(X=X,Y=Y,n.trees=100,shrinkage=.01,interaction.depth=1,
       tree.i <- finaltree[[m]][i]
       rel.infl[,m,i] <- ri.one(tree.i,n.trees=1,xnames)
       ## 1.3 Replace mth outcome with its residual, compute covariance           
-      Rm <- D 
+      Rm <- D
       Rm[,m] <- D[,m]-Dpred[,m,i]
-      Res.cov[,,m,i] <- cov(as.matrix(Rm),use="pairwise.complete")            
+      Res.cov[,,m,i] <- cov(as.matrix(Rm),use="pairwise.complete") 
+      #Res.cov[,,m,i] <- cov(D) - cov(D[,m]-Dpred[,m,i])
       ## 2. Evaluate loss criteria on training sample. Covariance reduced, correlation reduced, uls, or gls
       wm.raw[i,m] <- eval.loss(Rm=as.matrix(Rm[s,]),D=as.matrix(D[s,]),alpha=alpha,type=cov.discrep)
     }              
@@ -218,15 +253,30 @@ mvtb <- function(X=X,Y=Y,n.trees=100,shrinkage=.01,interaction.depth=1,
       }
     }
     # compute the best mod, and which xs were selected
-    besty <- bestys[i] <- which.max(wm.raw[i,])           
-    bestxs[i] <- bestx <- which.max(rel.infl[,besty,i])    
+    #besty <- bestys[i] <- which.max(wm.raw[i,])           
+    ## as an approximation, choose the best x by relative influence in each col at iteration i
+    bestxs[i,] <- bestx <- apply(rel.infl[,,i,drop=F],2,function(col){which.max(col)})    
     
-    # compute the covariance reduced by the best predictor
-    Sd <- cov(D)-Res.cov[,,besty,i]
-    covex[bestx,] <- covex[bestx,] + Sd[lower.tri(Sd,diag=TRUE)]         
+    # compute the covariance reduced (explained) by the best predictor for each outcome
+    correc <- 1+(1-shrinkage)
+    for(k in 1:m) {
+      Sd <- cov(D)-Res.cov[,,k,i]
+      Sd[lower.tri(Sd)] <- Sd[lower.tri(Sd)]*correc
+      if(k > 1) {
+        if(bestx[k] == bestx[k-1]) {
+        # the covariance elements will be counted twice. Don't want this to happen.
+        # since the covariance elements at column (row) k are already counted, set them to zero for k+1
+        # at k = m, this will only update the variance
+          Sd[1:(k-1),] <- 0
+          Sd[,1:(k-1)] <- 0
+        }
+      }
+      covex[bestx[k],] <- covex[bestx[k],] + Sd[lower.tri(Sd,diag=TRUE)]     
+      
+    }
     
     #Rm <- D - Dpred[,,i]
-    D <- Rm
+    D <- Y - yhat[,,i+1] # i=1 is init (colMeans Y)
     
     ## 7. Compute best iteration criteria: training and test error, sum of residual covariace matrix, weighted sum of residual covariance matrix
     #trainerr[i] <- mean((as.matrix(Rm[s,]))^2,na.rm=TRUE)
@@ -236,16 +286,18 @@ mvtb <- function(X=X,Y=Y,n.trees=100,shrinkage=.01,interaction.depth=1,
     #  testerr[i] <- mean((as.matrix(Rm[-s,]))^2,na.rm=TRUE)
     #}
     
+    ## 8. Compute weighted influences
+    for(m in 1:k) {
+      #rel.infl[,m] <- relative.influence(models[[m]],n.trees=i,scale=FALSE)
+      w.rel.infl[,m,i] <- rel.infl[,m,i]*wm.rel[i,m]
+    }
+    
     if(final.iter) {
       i <- i-1 # don't count the last iteration if all weights are  0
       break;
     }
   }
-  ## 8. Compute weighted influences
-  for(m in 1:k) {
-    #rel.infl[,m] <- relative.influence(models[[m]],n.trees=i,scale=FALSE)
-    w.rel.infl[,m,i] <- rel.infl[,m,i]*wm.rel[i,m]
-  }
+ 
   
   best.trees <- list(best.testerr=which.min(testerr),best.cv=best.iters.cv,last=i)
 
@@ -255,8 +307,8 @@ mvtb <- function(X=X,Y=Y,n.trees=100,shrinkage=.01,interaction.depth=1,
   fl <- list(models=models, covex=covex,maxiter=i,best.trees=best.trees,
              rel.infl=rel.infl, w.rel.infl=w.rel.infl,params=params,
              trainerr=trainerr[1:i],testerr=testerr[1:i],cv.err=cv.err[1:i],
-             bestxs=bestxs,bestys=bestys,
-             resid=Rm,ocv=ocv,
+             bestxs=bestxs,
+             resid=D,ocv=ocv,
              wm.raw=matrix(wm.raw[1:i,,drop=FALSE],nrow=i,ncol=k),wm.rel=wm.rel[1:i,,drop=FALSE],
              s=s,n=nrow(X),xnames=colnames(X),ynames=colnames(Y))
   if(compress) {
@@ -289,7 +341,7 @@ mvtb.fit <- function(X,Y,n.trees=100,shrinkage=.01,interaction.depth=1,
                               bag.fraction=bag.frac,keep.data=FALSE,...)
     }
     ## Multivariate mse. clumsy, I know.
-    yhat <- predict.mvtb(list(models=models),n.trees=1:n.trees,newdata=X)
+    yhat <- predict.mvtb(list(models=models),n.trees=1:n.trees,newdata=X,drop=FALSE)
     testerr <- trainerr <- rep(0,length=n.trees)
     for(i in 1:n.trees){
       R <- Y-yhat[,,i]
@@ -405,6 +457,7 @@ mvtbCV <- function(params) {
     # Last fold contains the full sample
     if(params$mc.cores > 1) {
         cores <- params$mc.cores
+        #cluster <- parallel::makeCluster(cores)
         params$mc.cores <- NULL
         out.k <- parallel::mclapply(1:(cv.folds+1),runone,params=params,cv.groups=cv.groups,sorig=sorig,mc.cores=cores)
     } else {
@@ -464,15 +517,17 @@ mvtbCV <- function(params) {
 #' Predicted values
 #' @param object mvtb object
 #' @param newdata matrix of predictors. 
-#' @param n.trees number of trees. If a list, returns predictions in an array. Defaults to the minimum best tree estimate.
+#' @param n.trees number of trees. If a vector, returns predictions in an array. Defaults to the minimum best tree estimate.
+#' @param drop Drop any dimensions of length 1.
 #' @param ... unused
-#' @return Returns an array of predictions for all outcomes, where the third dimension corresponds to the 
-#' predictions at a given number of trees.
+#' @return Returns an (array, matrix, vector) of predictions for all outcomes. The third dimension corresponds to the 
+#' predictions at a given number of trees, the second dimension corresponds to the number of outcomes.
 #' @export
-predict.mvtb <- function(object, n.trees=NULL, newdata, ...) {
+#' 
+predict.mvtb <- function(object, n.trees=NULL, newdata, drop=TRUE,...) {
   out <- object
   if(any(unlist(lapply(out,function(li){is.raw(li)})))){
-    out <- uncomp.mvtb(out)
+    out <- mvtb.uncomp(out)
   }
   if(is.null(n.trees)) { n.trees <- min(unlist(out$best.trees)) }
   K <- length(out$models)
@@ -486,5 +541,14 @@ predict.mvtb <- function(object, n.trees=NULL, newdata, ...) {
   #if(length(n.trees) == 1) {
   #  Pred <- drop(Pred)
   #}
+  if(drop){
+    Pred <- drop(Pred)
+  }
   return(Pred)
 }
+
+#residuals.mvtb <- fucntion(object,n.trees,...){
+#  
+#}
+
+
