@@ -1,3 +1,5 @@
+context("test_plot")
+
 set.seed(123)
 n <- 1000
 B <- matrix(0,nrow=5,ncol=4)
@@ -16,4 +18,7 @@ Y <- Xf %*% B + E
 out <- mvtb(Y=Y,X=X,n.trees=100,shrinkage = .5)
 plot(out)
 mvtb.perspec(out)
-heat.covex(out)
+mvtb.heat(out)
+mvtb.heat(out$covex)
+mvtb.heat(t(mvtb.ri(out)))
+mvtb.heat(t(mvtb.ri(out)),clust.method=NULL)
